!****************************************************************************** 
!+
! Module: omimagealign_mod
!
! Authors:  Christopher Brindle - MSSL
!         
!          
! Description: Contains various routines for aligning the RA/DEC
! coordinate axes of a list of source-list files
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
   
!
!**********************************************************************************
module omimagealign_mod
  use param
  use cal
  use dal
  use Errorhandling
  use sla
  use omallocation_mod
  use omconstants_mod
  use omastromcorrect_mod
  use omutility_mod
  use OmSourceParam_mod    

  implicit none 

  public


!******************************************************
! Structure used in source-matching
! x, y and z will store the equatorial rectangular
! coordinates of a source
! The distance between 2 sources is given by
! theta=acos(x1*x2 + y1*y2 + z1*z2),
! ie a.b=abcos(theta)
!
!*****************************************************
type ercoords

    real(double) :: rightAscension = 0_double
    real(double) :: declination = 0_double
    real(double) :: cosDec = 0.0_double
    real(double) :: x = 0_double     
    real(double) :: y = 0_double      
    real(double) :: z = 0_double
    real(single) :: poserr = 0.0
    real(single) :: significance = 0.0
    real(single) :: rawCountRate = 0.0 
    real(single) :: semiA = 0.0
    integer(int16) :: imin = 0
    integer(int16) :: qualityFlag = 0
    integer(int8) :: confusionFlag = 0

    integer(int32) :: number 
    integer(int32) :: noSourcePixels = 0 
    integer(int8) :: extensionFlag   = 0   
    type(Matcheddata) :: matches
end type ercoords

!*********************************************
!
!
!********************************************
type ExposureList

    integer :: noFiles = 0
    character(4) :: expId = ""
    character(4) :: filter = ""
    character(256), pointer :: fileName(:) => null()
    integer, pointer :: fileIndex(:) => null()
    logical, pointer :: aligned(:) => null()
   
end type ExposureList

!*********************************************
!
!
!********************************************
type WindowParameters

    integer :: windowX0 = 0
    integer :: windowDx = 0
    integer :: windowY0 = 0
    integer :: windowDy = 0
    logical :: binBpe = .false.

end type WindowParameters
!**********************************************************
!
!
!*********************************************************
type WindowList

    integer :: noFiles = 0
    character(4) :: expId = ""
  
    character(256), pointer :: fileList(:) => null()
    integer, pointer :: fileIndex(:) => null()
    character(5), pointer :: filterList(:) => null()  
    character(4), pointer :: exposureIdList(:) => null()  


    logical, pointer :: aligned(:) => null()
    real(single), pointer :: raOffsetList(:) => null()
    real(single), pointer :: decOffsetList(:) => null()
    real(single), pointer :: decScxList(:) => null()

    type(WindowParameters) :: window   

end type WindowList



interface operator(==)
    module procedure equalWindowParameters
end interface

interface computeNumberNeighbours
    module procedure computeNumberNeighbours1, &
                     computeNumberNeighbours2
end interface


!character(200), public :: fileName(1000)

contains 

!*******************************************************
! AddErrors
!
!*******************************************************
real(kind=double) function AddErrors1(error1, error2)
    real(double), intent(in) :: error1, error2
 
    AddErrors1 = dsqrt(error1 * error1  + error2 * error2)

end function AddErrors1


!*************************************************************
!
!************************************************************

logical function equalWindowParameters(win1, win2)
    type(windowParameters), intent(in) :: win1, win2

    if ( abs(win1%windowX0 - win2%windowX0) < 5 .and. &
         abs(win1%windowDX - win2%windowDX) < 5 .and. &
         abs(win1%windowY0 - win2%windowY0) < 5 .and. &
         abs(win1%windowDY - win2%windowDY) < 5 .and. &
        ( (win1%binBPE  .and. win2%binBPE) .or. & 
          ( .not. win1%binBPE  .and. .not. win2%binBPE)) )then  

       equalWindowParameters = .true.
    else
      equalWindowParameters = .false.
    end if
  
end function equalWindowParameters


!**********************************************************************
!
! 
!
!**********************************************************************

subroutine GetExposureList( nosw, fileList, noExposures, &
                            nrows, filterList, expList, exposureIdOutList )

    integer, intent(in) :: nosw
    character(*) :: fileList(:)
    integer, intent(out) :: noExposures
    integer, pointer :: nrows(:)
    character(*), pointer :: filterList(:)
    type(ExposureList), pointer :: expList(:)
    character(*), pointer :: exposureIdOutList(:)


    type(DataSetT) :: dataSetH
    integer :: i, j, k, n
    character(4) :: filter, exposureid
    character(4), pointer :: exposureIdList(:) => null()
    integer, pointer :: noFiles(:) => null()
   

    allocate(exposureIdList (1 : nosw))

    exposureIDList = ""
    noExposures = 0

    noFiles => allocateInt32(1, nosw, "noFiles")
    noFiles = 0

    !*************************************************************************
    ! Finds out the number of exposures
    !************************************************************************/

 iloop :  do i = 1, nosw
        dataSetH = dataSet( trim(fileList(i)), read)
        filter = trim( stringAttribute( dataSetH, "FILTER") )
        if( .not. hasAttribute( dataSetH, "EXPIDSTR" ) ) then

            call release(dataSetH)
            cycle
        end if
        exposureid = trim( stringAttribute( dataSetH, "EXPIDSTR") )
        call release(dataSetH)
  
        do j = 1, noExposures
            if ( exposureIDList(j) .eq. exposureID ) then
                nofiles(j) = noFiles(j) + 1 
                cycle iLoop
            end if
        end do
    
        !**************************************************
        ! Found a new exposure
        !************************************************** 
        noExposures = noExposures + 1        
        exposureIdList(noExposures) = exposureId 
        noFiles(noExposures) = 1

    end do iLoop


    if(noExposures .lt. 2) then

        deallocate(exposureIdList, nofiles)
        return
    end if

    !**************************************************************
    ! Fill out the exposure list
    !**************************************************************
    
    allocate(expList(1 : noExposures ))


    n = 0
    do i = 1, noExposures

        
        allocate( expList(i)%fileName(1 : noFiles(i)) )
        allocate( expList(i)%fileIndex(1 : noFiles(i)) )

        expList(i)%noFiles = noFiles(i) 
        expList(i)%expId = exposureIdList( i ) 
 
        k = 0       

        do j = 1, nosw
            dataSetH = dataSet( trim(fileList(j)), read)
            
            exposureid = trim( stringAttribute( dataSetH, "EXPIDSTR") )
            if (exposureid .eq. exposureIdList(i)) then
                k = k + 1
               
                expList(i)%fileName(k) = trim(fileList(j))
                expList(i)%fileIndex(k) = j

                n = n + 1
                nrows(n) = numberOfRows(table(dataSetH, "SRCLIST" ))



                filterList(n) = trim( stringAttribute( dataSetH, "FILTER") )                 
                exposureIdOutList(n) = exposureId                                
            end if
            call release(dataSetH)
        end do


        
!        write(*, *) "Exposure ", exposureIdList(i), " files-"
        
!        do j = 1, expList(i)%noFiles
!            write(*, *) j, " ) ",   trim(expList(i)%fileName(j))
!        end do
    end do

    n = 0



    do i = 1, noExposures
        
        do j = 1, expList(i)%noFiles
            n = n + 1
            fileList(n) = expList(i)%fileName(j)
        end do   
    end do
    deallocate(exposureIdList, nofiles)


end subroutine GetExposureList



!**********************************************************************
!
! Finds the number of "windows" and returns
! a list of parameters for each one 
! (WINDOWX0, WINDOWDX, WINDOWY), WINDOWDY and BINBPE)
!
!
!**********************************************************************
subroutine GetWindowList( nosw, fileList, filterList, noWindows, winList)

    integer, intent(in) :: nosw
    character(*), pointer :: fileList(:), filterList(:)
    integer, intent(out) :: noWindows

    type(WindowList), pointer :: winList(:)
  
    !*******************************************************
    ! Local variables
    !*******************************************************

    type(DataSetT) :: dataSetH
    type(TableT)   :: tableH
    integer :: i, j, k, l1, l2
    character(4) :: exposureid
    type(windowParameters), pointer :: winParameters(:) => null()
    integer, pointer :: noFiles(:) => null()
    character(256) :: name   
    type(WindowParameters) :: winData

    allocate(winParameters (1 : nosw))

    noWindows = 0

    noFiles => allocateInt32(1, nosw, "noFiles")
    noFiles = 0

 iloop :  do i = 1, nosw ! Loop through all the source-list files


        !********************************************
        ! Get the window parameters for this file
        !*******************************************
        dataSetH = dataSet( trim(fileList(i)), read)
        winData%windowX0 = int32Attribute( dataSetH, "WINDOWX0") 
        winData%windowDX = int32Attribute( dataSetH, "WINDOWDX") 
        winData%windowY0 = int32Attribute( dataSetH, "WINDOWY0") 
        winData%windowDY = int32Attribute( dataSetH, "WINDOWDY") 
        winData%binBPE = booleanAttribute( dataSetH, "BINBPE")      
        call release(dataSetH)
         
        !**********************************************************************
        ! Check to see if this window already in the list
        !*********************************************************************
        do j = 1, noWindows
            if ( winParameters(j) .eq. winData ) then
                nofiles(j) = noFiles(j) + 1 
                cycle iLoop      ! Yes- window already found
            end if
        end do

        !******************************************************************
        ! No - new window - add it to the list
        !******************************************************************
        noWindows = noWindows + 1        
        winParameters(noWindows) = winData 
        noFiles(noWindows) = 1

    end do iLoop

  
    !*************************************************************
    ! Fill up the window list
    !************************************************************
  
    allocate(winList(1 : noWindows ))


!write(*, *) "Omimagealign line 375 nowindows=", noWindows  

    do i = 1, noWindows ! Loop through all the windows
             
        allocate( winList(i)%fileList(1 : noFiles(i)) )
        allocate( winList(i)%fileIndex(1 : noFiles(i)) )
        allocate( winList(i)%filterList(1 : noFiles(i)) )
        allocate( winList(i)%exposureIdList(1 : noFiles(i)) )
        allocate( winList(i)%raOffsetList(1 : noFiles(i)) )
        allocate( winList(i)%decOffsetList(1 : noFiles(i)) )
        allocate( winList(i)%decScxList(1 : noFiles(i)) )

        winList(i)%window = winParameters(i)
        winList(i)%noFiles = noFiles(i) 
     
        k = 0       

        do j = 1, nosw ! Loop through all the files 
            dataSetH = dataSet( trim(fileList(j)), read)
            winData%windowX0 = int32Attribute( dataSetH, "WINDOWX0") 
            winData%windowDX = int32Attribute( dataSetH, "WINDOWDX") 
            winData%windowY0 = int32Attribute( dataSetH, "WINDOWY0") 
            winData%windowDY = int32Attribute( dataSetH, "WINDOWDY") 
            winData%binBPE = booleanAttribute( dataSetH, "BINBPE") 
            if( hasAttribute( dataSetH, "EXPOSUREID" )) then             
                exposureId = stringAttribute( dataSetH, "EXPOSUREID") 
            else
                write(exposureId, '(a1, i03)') 'S', j
            end if 
!        write(*, *) j, ") ", trim(fileList(j)), " x0=", winData%windowX0
          

               
            if (windata .eq. winList(i)%window) then
                k = k + 1

                name = trim(fileList(j))
                l1 = scan( name, '/', BACK=.true. ) + scan( name, '\', BACK=.true. ) + 1
       
                l2 = len( name )
  
                winList(i)%fileList(k) = name(l1 : l2)
                winList(i)%fileIndex(k) = j
                winList(i)%filterList(k) = filterList(j)
                winList(i)%exposureIdList(k) = exposureId

                winList(i)%window = winData

                winList(i)%decScxList(k) = real32Attribute( dataSetH, "DEC_PNT")
                tableH = Table(dataSetH, "SRCLIST")
                if (hasAttribute(tableH, "RAOFFSET")) then           
                    winList(i)%raOffsetList(k)  = real32Attribute(tableH, "RAOFFSET")
                    winList(i)%decOffsetList(k) = real32Attribute(tableH, "DEOFFSET")
                else
                    winList(i)%raOffsetList(k)  = -999.9
                    winList(i)%decOffsetList(k) = -999.9
                end if 
                                
                                              
            end if

            call release(dataSetH) ! Finished with the file

        end do
            
    end do

    deallocate(winParameters, nofiles)

end subroutine GetWindowList
!***************************************************************
!
! 1) For each window, the routine attempts to align the
!    source-list RA/DEC coordinates with the file that has
!    the most sources for the window.
!
! 2) Having corrected the coordinates in each window, the routine
!    will then attempt to align the coordinates in each window. 
!
!
!
!**************************************************************
subroutine AlignWindows(dx, source, nosw, nrows, winList, omFilter)
    integer(int8), intent(in) :: dx
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nosw
    integer, pointer :: nrows(:)
    type(windowList), pointer :: winList(:)
    character(*), intent(in) :: omFilter(:)

    type(sourceParameters), pointer, dimension(:) :: source1 => null(), source2 => null()
 
    integer :: i, j, k, k1, k2, n, n1, n2, noWindows, noFiles
    integer, pointer, dimension(:) :: nStart => null(), nFinish => null()
    logical, pointer :: success(:) => null()
    logical :: success1  
    real(double) ::meanXoffset, meanYOffset, xOffset1, yOffset1
    real(double), pointer, dimension(:) :: xoffset => null(), yoffset => null()

    real(double) :: minXOffset, minYOffset, maxXOffset, maxYOffset
    integer :: iStart, iFinish, jStart, jFinish, noOffsets, noIntervals
    integer :: noAligned
    integer :: noSuccesses, noFailures
    type(SHIFT), pointer, dimension(:) :: offset => null() 
    integer, pointer :: index(:) => null()
    logical, pointer :: aligned(:) => null()
    
  
    if (nosw .lt. 2) return

    nstart => allocateInt32(1, nosw, "nstart")
    nfinish => allocateInt32(1, nosw, "nfinish")
    
    nStart = 1
    nFinish = nrows(1)
    do i = 2, nosw
        nStart(i) = nFinish(i - 1) + 1
        nfinish(i) = nstart(i) + nrows(i) - 1
    end do
    
    nOWindows = size(winList)
    
    index => allocateInt32(1, noWindows, "index")

    write(errstr, '(a, a, i2, a)') &
            "Attempting to align the RA/DEC coordinate axes of the source-list files in each window", &
            " (Number of windows=",  noWindows, ")"
               
    call message(AppMsg, Verbose)

   
    iStart = -dx
    iFinish = dx
 
    jStart = -dx
    jFinish = dx

    if (minval(nrows) .ge. 5000) then
        noIntervals = 2
    else
        noIntervals = 4
    end if

    !******************************************************
    ! Set up the ra and dec offsets to be used in matching
    ! the sources with the catalogue stars
    !******************************************************
      
    noOffsets = ((abs(iStart) + abs(iFinish) + 1) * noIntervals) * &
               ((abs(jStart) + abs(jFinish) + 1) * noIntervals)  
    

    allocate( offset(1 : noOffsets) )


    call SetUpOffsets( noIntervals, istart, iFinish, jStart, jFinish, noOffsets, offset)
      

    !*************************************************************************
    ! For each window, align each file in the list
    ! with the file that has the most sources (hopefully an optical image)
    !**************************************************************************
 
    noFiles = 0

!*************************************
! FOR TEST
!***********************************
!source%rightAscension = PI
!source%declination = PI / 2.0

!    do i = 1, noWindows         

!        call random_number(xoffset1)
!        xoffset1 = (-1.0 + 2.0 * xoffset1) 

    

!       call random_number(yoffset1)
!       yoffset1 = (-1.0 + 2.0 * yoffset1) 

!     write(*, *) "Window ", i, " test xoffset=", xoffset1, " yoffset=", yoffset1 
 
!       xoffset1 = xoffset1 * degreesToRadians / 3600.0

!       yoffset1 = yoffset1 * degreesToRadians / 3600.0

      
!       do j = 1, winList(i)%noFiles
  
!           k2 = winList(i)%fileIndex(j)
            
!           source(nStart(k2) : nFinish(k2))%rightAscension = source(nStart(k2) : nFinish(k2))%rightAscension + xOffset1
!           source(nStart(k2) : nFinish(k2))%declination = source(nStart(k2) : nFinish(k2))%declination + yOffset1
!       end do

!    end do




!write(*, *) "Line 598  minval ra=", minval(source%rightAscension) * radiansToDegrees, &
!          " max=", maxval(source%rightAscension) * radiansToDegrees, " diff=", &
!           ( maxval(source%rightAscension) - minval(source%rightAscension) ) * radiansToDegrees * 3600.0

!write(*, *) "minval dec=", minval(source%declination) * radiansToDegrees, &
!            " max=", maxval(source%declination) * radiansToDegrees, " diff=", &
!       ( maxval(source%declination) - minval(source%declination) ) * radiansToDegrees * 3600.0
!write(*, *) "STARTING ALIGNMENT PROCESS"


    do i = 1, noWindows ! Loop through all the windows

!write(*, *) "Line 612 aligning sources in window ", i
!write(*, *) "minval ra=", minval(source%rightAscension) * radiansToDegrees, &
!          " max=", maxval(source%rightAscension) * radiansToDegrees, " diff=", &
!           ( maxval(source%rightAscension) - minval(source%rightAscension) ) * radiansToDegrees * 3600.0

!write(*, *) "minval dec=", minval(source%declination) * radiansToDegrees, &
!            " max=", maxval(source%declination) * radiansToDegrees, " diff=", &
!       ( maxval(source%declination) - minval(source%declination) ) * radiansToDegrees * 3600.0



        noFiles = noFiles + winList(i)%noFiles

        if (winlist(i)%noFiles .eq. 1) then
            index(i) = winlist(i)%fileindex(1)
            cycle
        end if

       
        meanXOffset = 0.0_double
        meanYOffset = 0.0_double
        minXOffset = 1000.0
        minYOffset = 1000.0
        maxXOffset = -1000.0
        maxYOffset = -1000.0


        xOffset => allocateDouble(1, winList(i)%noFiles, "xoffset")
        yOffset => allocateDouble(1, winList(i)%noFiles, "yoffset")
        xOffset = 0.0_double
        yOffset = 0.0_double

        !*********************************************************************
        ! Get the sources from the file in window i which has the maximum
        ! number of sources. Store in source1
        ! index(i) returns the file index and k the index the index in nosw
        ! k1 is the index in the nosw array- the input source-lists
        !  
        !**********************************************************************

        call getSources(winList(i), omFilter, source, nrows, &
                        nStart, nFinish, source1, index(i), k1)
        n1 = size(source1)

        write(errstr, '(a, i2, a, a, a, a, a, i3)') "Attempting to align window ", i, &
                   " images with file ", trim(winlist(i)%fileList(k1)), &
                   " (Filter=",  trim(winlist(i)%filterList(k1)),  &
                   " ) Number of files=", winlist(i)%noFiles                         

        call message(AppMsg, Verbose)
             


        !*********************************************************************
        ! Align all the files for this window with fileno index
        !********************************************************************

        noSuccesses = 0
        noFailures = 0
        allocate(success(1 : winList(i)%noFiles))
        success = .false.

        do j = 1, winList(i)%noFiles ! Loop through all the files for this window


  
            k2 = winList(i)%fileIndex(j)
!write(*, *) j, ") line 661 file=", trim(winList(i)%fileList(k2))
  


            if ( k2 .eq. index(i)) cycle ! Skip- same files


            n2 = nrows(k2)

  

            !*****************************************************
            ! Get the sources for this file
            !*****************************************************
            allocate(source2 (1 : n2 ))
            source2 = source(nStart(k2) : nFinish(k2)) 

         
!*******************************
! For test
!*******************************
!call random_number(num)
!if(num .gt. 0.5) then

!source2%rightAscension = source2%rightAscension + (1.0_double / 3600.0_double) * degreesToRadians
!source2%declination = source2%declination - (1.0_double / 3600.0_double) * degreesToRadians
!write(*, *) "xoffset=1.5,  y=-0.5"
!else   
!source2%rightAscension = source2%rightAscension - (1.0_double / 3600.0_double) * degreesToRadians
!source2%declination = source2%declination + (1.0_double / 3600.0_double) * degreesToRadians
!write(*, *) "xoffset=-1.5,  y=0.5"
!end if
!source(nStart(k2) : nFinish(k2)) = source2 


            !*******************************************************************************
            ! Find the best offsets to bring the RA/DEC coordinates of source1 and source2
            ! into agreement
            ! NB if xoffset/yoffset are positive then this means that the RAs/Decs
            ! of source1 are larger than those of source2
            ! ie source1 origin is to the right/top of source1 origin
            !********************************************************************************
            call AlignCoordinateAxes1(real(dx + 1), n2, source2, n1, source1, noOffsets, offset, &
                                      success(j), xoffset(j), yoffset(j))


            deallocate(source2)
            nullify(source2)

            if (success(j)) then ! Successful alignment


                !*****************************************************************************
                ! Correct the coordinates - add offsets to the coordinates
                ! xoffset and yoffset are added to the RA/DEC coordinates, respectively
                !***************************************************************************** 

                meanXoffset = meanXoffset + xoffset(j)
                meanYoffset = meanYoffset + yoffset(j)
                minXOffset = min( xOffset(j), minXoffset) 
                maxXOffset = max( xOffset(j), maxXoffset) 
                minYOffset = min( yOffset(j), minYoffset) 
                maxYOffset = max( yOffset(j), maxYoffset) 
                noSuccesses = noSuccesses + 1

!                write(errstr, '(a, i3, a, f6.2, a, f6.2, a, a, a)')  "File ", k, " aligned RA offset=",   &
!                     xoffset * radiansToDegrees * 3600.0_double, " Dec offset=",                          &
!                     yoffset * radiansToDegrees * 3600.0_double, " arcsecs (filter=",                     &
!                     trim(winlist(i)%filterList(j)), ")"                                         
!                call message(AppMsg, Verbose)
                       
                
            else
              
                noFailures = noFailures + 1

!                write(errstr, '(a, a, a, i5, a)' ) "Failed to align file ", &
!                       trim( winList(i)%fileList(j) ), &                              
!                       " (number of sources=", n2, ")"
!                call message(AppMsg, Verbose)
             
!                write(errstr, '(a, a, a, i5, a)' ) "with reference file ", &
!                       trim(winList(i)%fileList(index)), &
!                       " (number of sources=", n1, ")"           
!                call message(AppMsg, Verbose)


            end if
             

        end do  ! End of looping through the files in the window



        deallocate(source1)
        nullify(source1)
        write(errstr, '(a, i3, a, i3)') "Number of successes=", noSuccesses, " Number of failures=", noFailures              
        call message(AppMsg, Verbose)

!write(*, *) "Line 687"



        if(noSuccesses .gt. 0) then 
            meanXoffset = meanXOffset / dble(noSuccesses)
            meanYoffset = meanYoffset / dble(noSuccesses)
            minXoffset = minxoffset * radiansToDegrees * 3600.0_double
            maxXoffset = maxxoffset * radiansToDegrees * 3600.0_double
            minYoffset = minyoffset * radiansToDegrees * 3600.0_double
            maxYoffset = maxyoffset * radiansToDegrees * 3600.0_double
            write(errstr, '(a, f6.2, a, f6.2, a)') "Mean RA offset=", &
                           meanXOffset * radiansToDegrees * 3600.0_double, " Dec=",  &
                           meanYOffset * radiansToDegrees * 3600.0_double, " arcsecs"
            call message(AppMsg, Verbose)

            write(errstr, '(a, f6.2, a, f6.2, a)') "RA offset range=", minXOffset, "-",  maxXOffset, " arcsecs"
            call message(AppMsg, Verbose)
  

            write(errstr, '(a, f6.2, a, f6.2, a)') "Dec offset range=", minYOffset, "-",  maxYOffset, " arcsecs"
            call message(AppMsg, Verbose)
  
     
            !******************************************************************************
            !
            ! Correct the windows- first remove the mean
            !
            !******************************************************************************

! FOR test
!    meanxoffset = 0.0_double
!    meanyoffset = 0.0_double



            !*****************************************************************************
            ! Correct the coordinates - add offsets to the coordinates
            ! xoffset and yoffset are added to the RA/DEC coordinates, respectively
            !***************************************************************************** 
            call CorrectCoordinates(n1, source(nstart(index(i)) : nFinish(index(i)))%rightAscension,         &
                                            source(nstart(index(i)) : nfinish(index(i)))%declination,    &
                                            -meanXOffset, -meanYoffset)



            do j = 1, winList(i)%noFiles
  
                
                k = winList(i)%fileIndex(j)

                if ( k .eq. index(i)) cycle ! Skip- same files
                if (.not. success(j)) cycle


                n2 = nrows(k)
    
                xOffset(j) = xOffset(j) - meanXOffset
                yOffset(j) = yOffset(j) - meanYOffset

                !*****************************************************************************
                ! Correct the coordinates - add offsets to the coordinates
                ! xoffset and yoffset are added to the RA/DEC coordinates, respectively
                !***************************************************************************** 
                call CorrectCoordinates(n2, source(nstart(k) : nFinish(k))%rightAscension, &
                                            source(nstart(k) : nfinish(k))%declination,    &
                                            xoffset(j), yoffset(j))


            end do

        end if
        
        write(errstr, '(a, i3, a, i3)') "Number of successes=", noSuccesses, " failures=", noFailures                         
        call message(AppMsg, Verbose)

        deallocate(xOffset, yOffset, success)
        nullify(xOffset, yOffset, success)


    end do ! Loop through all the windows


!write(*, *) "Line 783 after aligning files in each window minval ra=", minval(source%rightAscension) * radiansToDegrees, &
!          " max=", maxval(source%rightAscension) * radiansToDegrees, " diff=", &
!           ( maxval(source%rightAscension) - minval(source%rightAscension) ) * radiansToDegrees * 3600.0
!write(*, *) "minval dec=", minval(source%declination) * radiansToDegrees, &
!            " max=", maxval(source%declination) * radiansToDegrees, " diff=", &
!       ( maxval(source%declination) - minval(source%declination) ) * radiansToDegrees * 3600.0


    if (noWindows .eq. 1) then

        if (associated(source1)) deallocate(source1)
        deallocate(nStart, nFinish, offset, index)
        nullify(source1, nStart, nFinish, offset, index)    
    
        return
    end if


    !******************************************************************
    !
    ! Now need to bring the individual windows into alignment
    !
    !*****************************************************************
    write(errstr, '(a, i2, a)') "Attempting to align windows (", noWindows, ")"                         
    call message(AppMsg, Verbose)
             

    allocate(aligned(1 : noWindows))
    aligned = .false.    ! No windows have been aligned
   
   
    noAligned = 0
   
    do i = 1, noWindows
     
        k1 = index(i)   ! The file number that the other images in the window were aligned with
!write(*, *) "Line 758 i=", i, " k1=", k1, " nrows=", nrows(k1)


        n1 = nrows(k1)  ! The number of sources in this file
        allocate(source1 (1 : n1 ))
        source1 = source(nStart(k1) : nFinish(k1))
          
        !************************************************************
        !
        ! Try to align windows i + 1 to noWindows with this window
        !
        !*************************************************************
        do j = i + 1, noWindows
            if(aligned(i) .and. aligned(j)) cycle    ! Both windows aligned

!write(*, *) "Line 772 trying to align window ", i, " with window ", j

            k2 = index(j)   ! The file number that the other images in the window were aligned with       
            n2 = nrows(k2)    

!write(*, *) "k2=", k2, " nrows=", n2


            allocate(source2 (1 : n2 ))
            source2 = source(nStart(k2) : nFinish(k2)) 
        
            !*******************************************************************************
            ! Find the best offsets to bring the RA/DEC coordinates of source1 and source2
            ! into agreement
            ! NB if xoffset/yoffset are positive then this means that the RAs/Decs
            ! of source1 are larger than those of source2
            ! ie source1 origin is to the right/top of source1 origin
            ! ie you add xoffset/yoffset to the RAs/Decs of source2 to bring RA/DEC axes
            ! into alignment
            !******************************************************************************** 
            call AlignCoordinateAxes1(real(dx + 1), n2, source2, n1, source1, noOffsets, &
                                      offset, success1, xoffset1, yoffset1)


            deallocate(source2)
            nullify(source2)
            if ( success1) then


!                        write(*, *) "Aligned window ", i, " with window ", j, &
!                       " xoffset=", (xoffset * 3600.0 * radiansToDegrees),    &
!                       " yoffset=", (yoffset * 3600.0 * radiansToDegrees)
                   
             

                !*************************************************************
                ! Shift all images of window j to bring into alignment with
                ! window i
                ! xoffset is added to RA, yoffset to Dec
                !***********************************************************
                if ( (.not. aligned(i) .and. .not. aligned(j)) .or. aligned(i)) then
!                write(*, *) "Aligned window ", j, " with window ", i
  

                    do n = 1, winlist(j)%noFiles
                        k = winList(j)%fileIndex( n )

                        !***************************************************************
                        ! If neither window i or window j have been aligned, or just
                        ! window i, then
                        ! align window j with window i
                        ! Add offsets
                        !
                        !**************************************************************                                                 
                        call  CorrectCoordinates(nrows(k), &
                              source(nstart(k) : nFinish(k))%rightAscension, &
                              source(nstart(k) : nfinish(k))%declination,    &
                              xoffset1, yoffset1)
                    end do
                else
!                   write(*, *) "*** Aligned window ***", i, " with window ", j
     

                    do n = 1, winlist(i)%noFiles
                        k = winList(i)%fileIndex( n )

!                        write(*, *) k, ") file=", winList(i)%fileList(n), " nrows=", nrows(k) 
                        !***************************************************************
                        ! 
                        ! align window i with window j
                        ! NB need to subtract the offsets from window i coordinates
                        !**************************************************************              
                        call  CorrectCoordinates(nrows(k), &
                              source(nstart(k) : nFinish(k))%rightAscension, &
                              source(nstart(k) : nfinish(k))%declination,    &
                              -xoffset1, -yoffset1)
           
                    end do
                    source1 = source(nStart(k1) : nFinish(k1))   

                end if
                aligned(i) = .true.
                aligned(j) = .true.
                noAligned = noAligned + 1

            end if
 
        end do  ! End of j loop 
        deallocate(source1)
        nullify(source1)
        
    end do ! End of i loop- aligning windows




!write(*, *) "Line 943 after aligning windows minval ra=", minval(source%rightAscension) * radiansToDegrees, &
!          " max=", maxval(source%rightAscension) * radiansToDegrees, &
!          " diff=",  ( maxval(source%rightascension) - minval(source%rightascension) ) * radiansToDegrees * 3600.0

!write(*, *) "minval dec=", minval(source%declination) * radiansToDegrees, &
!  " max=", maxval(source%declination) * radiansToDegrees, &
!  " diff=",  ( maxval(source%declination) - minval(source%declination) ) * radiansToDegrees * 3600.0




    write(errstr, '(a, i2)') "Number of windows aligned=", noAligned                         
    call message(AppMsg, Verbose)
             



    if (associated(source1)) deallocate(source1)
    deallocate(aligned, nStart, nFinish, index)
    deallocate(offset)
    nullify(source1, nStart, nFinish, offset, index)    
    
end subroutine AlignWindows


!*************************************************
! Tries to compute the shifts, xoffset and yoffset 
! that need to be applied to align the RA/DEC axes
! of the sources in source1 and source2
! NB xOffset = sum(ra2 - ra1) / n
!    yOffset = sum(dec2 - dec1) / n
!
! if successful then success is set to true, otherwise false
!  
!************************************************
Subroutine AlignCoordinateAxes1(tol, n1, source1In, n2, &
                source2In, noOffsets, offset, &
                success, xoffset, yoffset)
    real(single), intent(in) :: tol
    integer, intent(in) :: n1, n2
    type(sourceParameters) :: source1In(:), source2In(:)
    integer, intent(in) :: noOffsets  
    type(shift), pointer :: offset(:)
    logical, intent(out) :: success
    real(double), intent(out) :: xoffset, yoffset  
    integer, pointer :: sortIndex(:) => null()

    !****************************************************************************
    ! Local variables
    !****************************************************************************
    integer(int32) :: i, l, k, m, m1, kMerit
    integer(int32) :: nmatched, maxnmatched, minnmatched
    real(double) :: dx, dy
    real(double) :: sum1, sum2, sum3, sum4, sum5, sum6, diff, error
    
   
    type(ercoords) :: tempSource
    real(double) :: meanRadius, meanRadiusMin, weight, sumWeight 
    real(single) :: minRadius, tolerance
    integer :: nCount, nCat, minCount
    real(single) :: merit, minMerit 
    !**************************************************************************
    ! Do the source-matching using equatorial rectangular coordinates
    !**************************************************************************
    type(ercoords), pointer, dimension(:) :: source1 => null(), source2 => null()


  
    allocate( source1(1 : n1) )
 
    do i = 1, n1
        source1(i)%rightAscension = source1in(i)%rightAscension
        source1(i)%declination    = source1in(i)%declination
       

        source1(i)%poserr         = source1In(i)%poserr
        source1(i)%extensionFlag  = source1In(i)%extensionFlag
        source1(i)%qualityFlag  = source1In(i)%qualityFlag
        source1(i)%confusionFlag  = source1In(i)%confusionFlag


        source1(i)%significance   = source1In(i)%significance

        source1(i)%rawCountRate   = source1In(i)%RawCountRate

        source1(i)%poserr = max(source1(i)%poserr, 0.5)   

     
        if ( btest(source1(i)%qualityFlag, 4) ) then
            source1(i)%poserr = max(source1(i)%poserr, 1.0)   
        end if
        !***************************************************************
        ! Compute the rectangular coordinates for this source
        !***************************************************************
        call  computeRectangularCoordinates3( source1(i) )

    end do


    !********************************************************
    !
    ! Produce a new list of catalogue stars
    !*******************************************************
    allocate( source2(1 : n2) )
 
 
    do i = 1, n2
        source2(i)%rightAscension = source2in(i)%rightAscension
        source2(i)%declination    = source2in(i)%declination
        source2(i)%poserr         = source2In(i)%poserr
        source2(i)%extensionFlag  = source2In(i)%extensionFlag
        source2(i)%qualityFlag    = source2In(i)%qualityFlag
        source2(i)%confusionFlag    = source2In(i)%confusionFlag
        source2(i)%significance   = source2In(i)%significance       
        source2(i)%rawCountRate   = source2In(i)%rawCountRate
 
     
        source2(i)%poserr = max(source2In(i)%poserr, 0.5)   

        if ( btest(source2(i)%qualityFlag, 4) ) then
            source2(i)%poserr = max(source2(i)%poserr, 1.0)   
        end if

        !***************************************************************
        ! Compute the rectangular coordinates for this source
        ! NB Coordinates are in radians
        !***************************************************************

        call  computeRectangularCoordinates3( source2(i) )

    end do

    
    call computeNumberNeighbours2(source1, source2, tol, 10_int16)

    call computeNumberNeighbours2(source2, source1, tol, 10_int16)
       
    !************************************************
    !
    ! Check overlap of stars with catalogue stars
    !
    !************************************************
    nCount = 0

    do i = 1, n1


        if(source1(i)%matches%numMatches .eq. 0) cycle
        if (source1(i)%extensionFlag .eq. 1) cycle
        if (source1(i)%qualityFlag .gt. 0) cycle
        if (source1(i)%confusionFlag .gt. 0) cycle
 
        if(source1(i)%significance < 3.0) cycle ! Ignore very faint sources  
        if(source1(i)%rawCountRate > 50.0) cycle ! Ignore very bright sources  
                     
        nCount = nCount + 1
        source1(ncount) = source1(i)
       
    end do


    !*****************************************************
    ! Match the sources in order of decreasing significance
    !****************************************************
    sortIndex => allocateInt32(1, ncount, "sortIndex")
    call IndexSort(source1%significance, sortIndex, ncount)

    nCat = 0

    do i = 1, n2   ! The number of source2 sources 


        if(source2(i)%matches%numMatches .eq. 0) cycle
        if (source2(i)%extensionFlag .eq. 1) cycle 
        if (source2(i)%qualityFlag .gt. 0) cycle
        if (source2(i)%confusionFlag .gt. 0) cycle
 

        if(source2(i)%significance < 3.0) cycle ! Ignore very faint sources  
        if(source2(i)%rawCountRate > 50.0) cycle ! Ignore very bright sources  
               
     

        nCat = nCat + 1
        source2(ncat) = source2(i)
        
    end do


    call computeNumberNeighbours2(source1(1 : nCount), source2(1 : ncat), tolerance, 10_int16)
    call computeNumberNeighbours2(source2(1 : ncat), source1(1 : nCount), tolerance, 10_int16)

   
    mincount = max(4, min(ncat, ncount) / 2)  

 
    if ( nCount .lt. mincount ) then
!write(*, *) "Line 1190 mincount=", mincount, " ncount=", ncount, " ncat=", ncat
!        write(errstr,'(a, i5, a)') "Too few stars (", nCount , ") overlap with catalogue stars for alignment attempt"
!        call message(AppMsg, Verbose)
        deallocate(source1, source2)
        success = .false.
        return
    end if 
   
    xoffset = 0.0
    yoffset = 0.0


    meanRadiusMin = 10.0 ! The minimum weighted sum of source separations
         

    maxnmatched = 3
    minnmatched = 3

    tolerance = 2.0


    !***************************************************************
    ! Loop through all the offsets
    !***************************************************************
    minMerit = 0.5
   
    do i = 1, noOffsets ! Loop through all the x offsets
        
        dx = offset(i)%x / max(0.1_double, source1(1)%cosdec)               
        dy = offset(i)%y
 
        nmatched = 0
        meanRadius = 0.0_double
        sumWeight = 0.0_double
               
        sum1 = 0.0_double
        sum2 = 0.0_double
        sum3 = 0.0_double
        sum4 = 0.0_double
        sum5 = 0.0_double
        sum6 = 0.0_double

        kMerit = 0         

        do k = ncount, 1, -1 ! Loop through the sources of file l and match them with the sources in file maxfile
                       
            l = sortIndex(k)  ! Match in order of decreasing source significance                  

   
            tempSource = source1(l)
           

            tempSource%rightAscension = tempSource%rightAscension + dx ! Add x offset to x coordinate                       
            tempSource%declination = tempSource%declination + dy       ! Add y offset to y coordinate

            !*******************************************************************
            ! Compute the equatorial rectangular coordinates 
            !*******************************************************************
            call computeRectangularCoordinates3( tempSource )    

            !*******************************************************************
            !
            ! Try to find a match to tempsource
            !
            !******************************************************************

            if(SourceMatch2(tempsource, source2,                    &
                            tolerance,                              &               
                            minRadius, m)) then

                error = max(0.5_double, AddErrors1(dble(tempSource%poserr), dble(source2(m)%poserr)))
                                
                weight =  1.0_double / (error * error)
                
                meanRadius = meanRadius + dble(minRadius) * weight
                sumWeight = sumWeight + weight                                                  
                nmatched = nmatched + 1 ! Number of source matched for the current offsets                 
                if (nMatched .gt. 1) then

                    if(source2(m)%significance .le. source2(m1)%significance ) then
                        kmerit = kmerit + 1
                    end if
                        
                end if
              
                m1 = m
                
                        
                sum1 = sum1 + weight * source1(l)%rightAscension
                sum2 = sum2 + weight * source1(l)%declination
                sum3 = sum3 + weight * source2(m)%rightAscension
                sum4 = sum4 + weight * source2(m)%declination
                diff = source2(m)%rightAscension - tempSource%rightAscension
                sum5 = sum5 + weight * diff

                diff = source2(m)%declination - tempSource%declination
                sum6 = sum6 + weight * diff


            end if

        end do

!write(*, *) "Line 1285 nmatched=", nMatched        
        if (sumWeight .gt. 0.0_double .and. nMatched .gt. 0) then
                
            merit = real(kMerit) / real(nmatched) 
            if (nmatched .ge. maxnmatched .and. merit .ge. minMerit) then   ! Save the best overlapping offsets
 
                meanRadius = meanRadius / sumWeight 
                if(meanRadius .gt. meanRadiusMin) cycle
                sum5 = (sum5 / sumWeight) * source1(1)%cosdec
                if(abs(sum5) .gt. 0.5) cycle 
                sum6 = sum6 / sumWeight
                if(abs(sum6) .gt. 0.5) cycle 
     
                maxnmatched = nmatched        
                minMerit = merit
                meanRadiusMin = meanRadius  

                xoffset = (sum3 - sum1) / sumWeight ! Means add xoffset to source1
                yoffset = (sum4 - sum2) / sumWeight                   

!write(*, *) "sum1=", sum1, " sum3=", sum3, " sumweight=", sumWeight, " xoffset=", xoffset * radiansToArcsecs
!write(*, *) "sum4=", sum4, " sum2=", sum2, " sumweight=", sumWeight, " yoffset=", yoffset * radiansToArcsecs
!write(*, *) "nmatched=", nmatched, " max matched=", maxnmatched, " dec=", source1(1)%declination * radiansToDegrees

                if (nmatched .eq. n1) exit ! Can't match any more                                  
            end if
        end if
          
    end do 

    if (maxnmatched .ge. minnmatched .and. meanRadiusMin .le. 3.0) then
        
            
        success = .true.

!        write(errstr, '(a, f5.2, a, f5.2, a)') "Least-squares offsets - RA offset=", &
!                       (xoffset * radiansToDegrees * 3600.0_double),                                   &
!                       " arcsecs, Dec offset=", (yoffset * radiansToDegrees * 3600.0_double), &
!                        " arcsecs"
!        call message(AppMsg, Verbose)

 
    else
        success = .false.  
    end if  
   

    deallocate(source1, source2)
    nullify(source1, source2)
  
end Subroutine AlignCoordinateAxes1



!*************************************************************
! Correct the coordinates using the offsets
! Everything shoud be in radians
!*************************************************************
subroutine CorrectCoordinates(nSource, ra, dec, xoffset, yoffset)
    integer, intent(in) :: nSource
    real(double) :: ra(:), dec(:)
    real(double), intent(in) :: xoffset, yoffset
    integer :: i

!write(*, *) "coorectcoordinates - xoffset=", xoffset * radiansToDegrees * 3600.0, &
! " yoffset=", yoffset * radiansToDegrees * 3600.0


! write(*, *) "before correcting minval(ra)=", minval(ra) * degreesToRadians, &
!            " max=", maxval(ra) * degreesToRadians, &
!             " minval(dec)=", minval(dec) * degreesToRadians, &
!             " max=", maxval(dec) * degreesToRadians


    do i = 1, nSource
        ra(i) = ra(i) + xoffset
        if(ra(i) .gt. TWO_PI) then 
            ra(i) = ra(i) - TWO_PI

        else if(ra(i) .lt. 0.0_double) then
            ra(i) = ra(i) + TWO_PI
        end if 

        dec(i) = dec(i) + yoffset
    end do




!write(*, *) "after correcting minval(ra)=", minval(ra) * degreesToRadians, &
!            " max=", maxval(ra) * degreesToRadians, &
!             " minval(dec)=", minval(dec) * degreesToRadians, &
!             " max=", maxval(dec) * degreesToRadians 
        




end subroutine CorrectCoordinates





!***************************************************
! Function SourceMatch2
! Tries to match a catalogue star to a given source
! source is the source to be matched
! catsource are the sources for source to be matched with
!
!***************************************************
logical function SourceMatch2(source,                    &
                              catSource,                 &
                              tolerance,                 &
                              minRadius,                 & 
                              matchedIndex )
 

    type(ercoords), intent(in) :: source
    type(ercoords), pointer :: catSource(:)   
    real(single), intent(in) :: tolerance
    real(single), intent(out) :: minRadius
    
    integer, intent(out) :: matchedIndex 
    

    real(single) :: error1  
    real(double) :: distance 

    integer(int32) :: i, j
 
   
    if (source%matches%numMatches .eq. 0) then
        SourceMatch2 = .false.
        return
    end if

   
    matchedIndex = -1
    minRadius = 1e9
    error1 = source%poserr

    do i = 1, source%matches%numMatches
      
        j = source%matches%index(i)  
          
        !********************************************************************
        ! Find the distance between the sources
        !*********************************************************************           
        distance = GreatCircleDistance(source%x, source%y, source%z, &
                                catSource(j)%x, catsource(j)%y, catsource(j)%z)
                         
        if (distance .le. tolerance) then ! Likely match to star i  
                 
            if ( distance .lt. minRadius) then 
                minRadius = distance
                matchedindex = j
            end if   
               
        end if

    end do
    
    if(matchedIndex > 0) then
        sourceMatch2 = .true.

!write(*, *) "Matched index=", matchedIndex, " minradius=", minradius
!write(*, *) "ra=", source%rightAscension, " dec=", source%declination
!write(*, *) "matched ra=", catsource(matchedIndex)%rightAscension, " dec=", &
!catsource(matchedIndex)%declination

    else
        sourceMatch2 = .false.       
    end if


end function SourceMatch2   

!***************************************************************
!

!**************************************************************

subroutine AlignRudi5Images(source, nosw, nrows, expList, exposureIdList)
    type(sourceParameters) :: source(:)
    integer, intent(in) :: nosw
    integer, pointer :: nrows(:)
    type(ExposureList), pointer :: expList(:)
    character(*), pointer :: exposureIdList(:)
    integer :: i  
    character(4), dimension(5, 6) :: exposureId

    !**********************************************************************
    !                   V        U       B      UVW1    UVM2    UVW2    
    !**********************************************************************
    exposureId(1, :) = (/"S006", "S007", "S008", "S010", "S011", "S012"/)
    exposureId(2, :) = (/"S485", "S489", "S493", "S501", "S505", "S509"/)
    exposureId(3, :) = (/"S486", "S490", "S494", "S502", "S506", "S510"/)
    exposureId(4, :) = (/"S487", "S491", "S495", "S503", "S507", "S511"/)
    exposureId(5, :) = (/"S488", "S492", "S496", "S504", "S508", "S512"/)

!    write(*, *) "Alignrudi5images line 1431" 
            
    do i = 1, 5
!        write(*, *) i, ")-"
        call AlignImages(source, nosw, nrows, expList, exposureId(i, :), exposureIdList)

    end do

end subroutine AlignRudi5Images





!***************************************************************
!
!
!
!**************************************************************

subroutine AlignImages(source, nosw, nrows, expList, exposureId, exposureIdList)

    type(sourceParameters) :: source(:)
    integer, intent(in) :: nosw
    integer, pointer :: nrows(:)
    type(ExposureList), pointer :: expList(:)
    

    character(4), intent(in) :: exposureId(6)
    character(*), pointer :: exposureIdList(:)
    integer :: nStart, nFinish, n1, n2, noExposures, index, i, j
    logical :: done
    type(sourceParameters), pointer, dimension(:) :: source1 => null(), source2 => null()
    real(double) :: xoffset, yoffset

    integer :: iStart, iFinish, jStart, jFinish, noOffsets, noIntervals

    type(SHIFT), pointer, dimension(:) :: offset => null() 

    nStart = 1
    nFinish = nrows(1)

    noExposures = size(expList)

    index = 0
iLoop : do i = 1, noExposures - 1

        do j = 1, 6
            
            if (explist(i)%expid .eq. exposureId(j)) then
                index = expList(i)%fileIndex(1)
                
                exit iLoop
            end if

        end do
    end do iLoop

!if (index .eq. 0) call fatal("bug at line 5391")

    if ( expList(index)%noFiles .eq. 2) index = index + 1


    !************************************
    !
    ! Try to match all exposures with this file
    !************************************


     if (index .gt. 1) then
         nStart = sum( nrows( 1 : index - 1)) - 1         
     else
         nStart = 1
     end if


     nFinish = nStart + nrows(index) - 1
     n1 = nrows(index)
     allocate(source1 (1 : n1 ))
     source1 = source(nStart : nFinish)
    

    iStart = -2
    iFinish = 2
 
    jStart = -2
    jFinish = 2

    if (minval(nrows) .ge. 5000) then
        noIntervals = 2
    else
        noIntervals = 4
    end if

    !******************************************************
    ! Set up the ra and dec offsets to be used in matching
    ! the sources with the catalogue stars
    !******************************************************
      
    noOffsets = ((abs(iStart) + abs(iFinish) + 1) * noIntervals) * &
               ((abs(jStart) + abs(jFinish) + 1) * noIntervals)  
    

    allocate( offset(1 : noOffsets) )


    call SetUpOffsets( noIntervals, istart, iFinish, jStart, jFinish, noOffsets, offset)
      
    
    do i = 1, nosw

        if (i .eq. index ) cycle
        if ( count(exposureId .eq. exposureIdList(i) ) .eq. 0) cycle
            

        if (i .gt. 1) then
            nStart = sum( nrows( 1 : i - 1)) - 1
         
        else
            nStart = 1
        end if
        nFinish = nStart + nrows(i) - 1
        n2 = nrows(i) 


        allocate(source2 (1 : n2 ))
        source2 = source(nStart : nFinish)
        
 
        call AlignCoordinateAxes1(2.0, n2, source2, n1, source1, noOffsets, offset, done, xoffset, yoffset)

            
        deallocate(source2)
        nullify(source2)
      

    end do
    
    deallocate(source1)
    nullify(source1)
      

end subroutine AlignImages


!**********************************************************************************************
! Computes the angular distance between 2 points on the celestial sphere, using
! the equatorial rectangular coordinates of the objects
!
!*********************************************************************************************

real(double) function greatCircleDistance( x1, y1, z1, x2, y2, z2)
    real(double), intent(in) :: x1, y1, z1, x2, y2, z2
    

    greatCircleDistance = acos( min(1.0_double, (x1 * x2 + &
                                          y1 * y2 + &
                                          z1 * z2)) ) * radiansToArcSeconds

end function greatCircleDistance
!********************************************************************************************
!
!*******************************************************************************************
subroutine computeRectangularCoordinates( source )
    type(sourceParameters), intent(inout) :: source

    real(double) :: cosDec

    cosDec = cos(source%declination * degreesToRadians)

    source%apertureRadius = cosDec * cos(source%rightAscension * degreesToRadians)
    source%axialRatio = cosDec * sin(source%rightAscension * degreesToRadians)
    source%alpha = sin(source%declination * degreesToRadians)

end subroutine computeRectangularCoordinates

!********************************************************************************************
!
! Computes rectangular cartesian coordinates from RA/DEC
!
!*******************************************************************************************
subroutine computeRectangularCoordinates3( source )
    type(ercoords), intent(inout) :: source

    real(double) :: cosDec



    cosDec = cos(source%declination  )
    source%cosDec = cosDec
    source%x = cosDec * cos(source%rightAscension )
    source%y = cosDec * sin(source%rightAscension  )
    source%z = sin(source%declination  )

end subroutine computeRectangularCoordinates3
!********************************************************************************************
!
! Computes rectangular cartesian coordinates from RA/DEC
!
!*******************************************************************************************
subroutine computeRectangularCoordinates1( source )
    type(ercoords), intent(inout) :: source

    real(double) :: cosDec

    cosDec = cos(source%declination * degreesToRadians )
    source%cosDec = cosDec
    source%x = cosDec * cos(source%rightAscension * degreesToRadians)
    source%y = cosDec * sin(source%rightAscension * degreesToRadians )
    source%z = sin(source%declination * degreesToRadians )

end subroutine computeRectangularCoordinates1

!********************************************************************************************
! Computes rectangular cartesian coordinates from RA/DEC
!
!*******************************************************************************************
subroutine computeRectangularCoordinates2( ra, dec, x, y, z)
    real(double), intent(in) :: ra, dec
    real(double), intent(inout) :: x, y, z

    real(double) :: cosDec

    cosDec = cos( dec )
    x = cosDec * cos(ra )
    y = cosDec * sin(ra )
    z = sin( dec )

end subroutine computeRectangularCoordinates2
!***************************************************************
! For each source, counts the number of sources from other
! files that lie within a distance tolerance
!
!***************************************************************
subroutine computeNumberNeighbours1(source, tolerance, maxSize)
    type(ercoords) :: source(:)
    real(single), intent(in) :: tolerance
    integer(int16), intent(in) :: maxSize
    integer :: i, j, n
    real(double) :: tol, dra, distance

    !*****************************************************************************
    ! masSize is the maximum number of "neighbours" that a source can have
    !
    !*****************************************************************************

    call setMaxNumMatches(maxSize)

!write(*, *) "ComputeNumberneighbours1 line 1859"

    tol = tolerance * degreesToRadians / 3600.0
    n = size(source)
    do i = 1, n
        
        if(i .gt. 1) then  
            do j = i - 1, 1, -1
                if(source(i)%imin .eq. source(j)%imin) cycle

                dra = abs(source(i)%rightAscension - source(j)%rightAscension)
                if (dra .lt. 6.2744586) then 
                    if( (dra * source(i)%cosDec) .gt. tol) exit
                end if
                if( abs(source(i)%declination - source(j)%declination) .gt. tol) cycle
               
                distance = GreatCircleDistance(source(i)%x, source(i)%y, source(i)%z, &
                                               source(j)%x, source(j)%y, source(j)%z)

                
                if(distance .le. tolerance) then 
                 
!                    write(*, *) "Adding to list indx ", j, " distance=", distance
        
                    call addToList(source(i)%matches, j, distance)
                end if
            end do
        end if

        if (i .lt. n) then
          
            do j = i + 1, n
                if(source(i)%imin .eq. source(j)%imin) cycle

                dra = abs(source(i)%rightAscension - source(j)%rightAscension)
                if (dra .lt. 6.2744586) then 
                    if( (dra * source(i)%cosDec) .gt. tol) exit
                end if
                if( abs(source(i)%declination - source(j)%declination) .gt. tol) cycle
               
                distance = GreatCircleDistance(source(i)%x, source(i)%y, source(i)%z, &
                                               source(j)%x, source(j)%y, source(j)%z)

        
                if(distance .le. tolerance) then 

!                    write(*, *) "Adding to list indx ", j, " distance=", distance
                    call addToList(source(i)%matches, j, distance)

            

                end if
     

             end do
        end if

!        write(*, *) i, ") source has ", source(i)%matches%numMatches, " neighbours" 
!       do j = 1, source(i)%matches%NumMatches
!           write(*, *) j, ") index=", source(i)%matches%index(j), " radius=", source(i)%matches%radius(j)
!      end do

    end do

!write(*, *) "computenumberneighbours1 maxsize=", maxSize

end subroutine computeNumberNeighbours1
!***************************************************************
! For each source, counts the number of sources from other
! files that lie within a distance tolerance
!
!***************************************************************
subroutine computeNumberNeighbours2(source1, source2, tolerance, maxSize)
    type(ercoords), dimension(:) :: source1, source2
    real(single), intent(in) :: tolerance
    integer(int16), intent(in) :: maxSize
    integer :: i, j, n1, n2
    real(double) :: tol, dra, distance

    call setMaxNumMatches( maxSize )

    tol = tolerance / 3600.0

    n1 = size(source1)
    n2 = size(source2)

    do i = 1, n1
     
        do j = 1, n2
           

            dra = abs(source1(i)%rightAscension - source2(j)%rightAscension)
            if (dra .lt. 359.99) then 
                if( (dra * source1(i)%cosDec) .gt. tol) cycle
            end if
            if( abs(source1(i)%declination - source2(j)%declination) .gt. tol) cycle
               
            distance = GreatCircleDistance(source1(i)%x, source1(i)%y, source1(i)%z, &
                                               source2(j)%x, source2(j)%y, source2(j)%z)

!                write(*, *) j, ") distance=", distance, & 
!            " ra=", source2(j)%rightAscension * radiansToDegrees, &  
!            " dec=", source2(j)%declination * radiansToDegrees  


            if(distance .le. tolerance) then 



!write(*, *) "Line 1826 adding to list source ", j, " distance=", distance, &
!            " ra=", source2(j)%rightAscension * radiansToDegrees, &  
!            " dec=", source2(j)%declination * radiansToDegrees, &
!            " x=", source2(j)%x, " y=", source2(j)%y, " z=", source2(j)%z  

                call addToList(source1(i)%matches, j, distance)

            end if
             
        end do
      
         
     
!        write(*, *) i, ") source has ", source1(i)%matches%numMatches, " neighbours", &
!            " ra=", source1(i)%rightAscension * radiansToDegrees, &  
!            " dec=", source1(i)%declination * radiansToDegrees, &
!            " x=", source1(i)%x, " y=", source1(i)%y, " z=", source1(i)%z  

    end do


!write(*, *) "tolerance=", tolerance, " tol=", tol
end subroutine computeNumberNeighbours2


!******************************************************************
!
! Find the file in the input  window that has the most sources
! and return a list of the sources
!
!*****************************************************************
subroutine getSources(winList, omFilter, source, nrows, &
           nStart, nFinish, source1, k, index)
        
    type(WindowList), intent(in) :: winList
    character(*), intent(in) :: omFilter(:)

    type(sourceParameters), pointer, dimension(:) :: source
    integer, pointer, dimension(:) :: nrows, nStart, nFinish
    type(sourceParameters), pointer, dimension(:) :: source1
    integer, intent(out) :: index, k

    integer :: j, l, n, maxCount

    if(associated(source1)) then
        deallocate(source1)

    end if
    maxCount = 0
    index = 1
    do j = 1, winlist%noFiles
       
        do k = size(omFilter), 1, -1
            if ( omFilter(k) .eq. winList%filterList(j) ) then
                l = winList%fileIndex(j)
                n = nrows(l)
                   
                if (n .gt. maxCount) then
                    maxcount = n                   
                    index = j
                end if
            end if
        end do 

    end do


    k = winlist%fileIndex(index) ! k is the index of the file in the file-list
    
    !******************************************************************
    ! Store the sources for this file in array source1
    !******************************************************************
    allocate(source1 (1 : nrows(k) ))
    source1 = source(nStart(k) : nFinish(k))
    

end subroutine getSources

end module omimagealign_mod


  








